@@ -66,6 +66,8 @@ typedef enum {
 	NAUTILUS_ICON_LABEL_POSITION_BESIDE
 } NautilusIconLabelPosition;
 
+#define	NAUTILUS_ICON_CONTAINER_TYPESELECT_FLUSH_DELAY 1000000
+
 typedef struct NautilusIconContainerDetails NautilusIconContainerDetails;
 
 typedef struct {
@@ -95,11 +97,22 @@ typedef struct {
 						   GList *item_uris,
 						   GdkPoint *relative_item_points,
 						   const char *target_uri,
-						   int copy_action,
+						   GdkDragAction action,
+						   int x,
+						   int y);
+	void	     (* handle_url)		  (NautilusIconContainer *container,
+						   char *url,
+						   GdkDragAction action,
 						   int x,
 						   int y);
 	void	     (* handle_uri_list)    	  (NautilusIconContainer *container,
-						   GList *item_uris,
+						   char *uri_list,
+						   GdkDragAction action,
+						   int x,
+						   int y);
+	void	     (* handle_text)		  (NautilusIconContainer *container,
+						   char *text,
+						   GdkDragAction action,
 						   int x,
 						   int y);
 
@@ -241,7 +254,8 @@ gboolean          nautilus_icon_container_has_stretch_handles           (Nautilu
 gboolean          nautilus_icon_container_is_stretched                  (NautilusIconContainer  *container);
 void              nautilus_icon_container_show_stretch_handles          (NautilusIconContainer  *container);
 void              nautilus_icon_container_unstretch                     (NautilusIconContainer  *container);
-void              nautilus_icon_container_start_renaming_selected_item  (NautilusIconContainer  *container);
+void              nautilus_icon_container_start_renaming_selected_item  (NautilusIconContainer  *container,
+									 gboolean                select_all);
 
 /* options */
 NautilusZoomLevel nautilus_icon_container_get_zoom_level                (NautilusIconContainer  *view);
