@@ -3,6 +3,7 @@
    nautilus-thumbnails.h: Thumbnail code for icon factory.
  
    Copyright (C) 2000, 2001 Eazel, Inc.
+   Copyright (C) 2002, 2003 Red Hat, Inc.
   
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -29,7 +30,6 @@
 #include "nautilus-global-preferences.h"
 #include "nautilus-icon-factory-private.h"
 #include "nautilus-icon-factory.h"
-#include "nautilus-theme.h"
 #include <eel/eel-gdk-pixbuf-extensions.h>
 #include <eel/eel-graphic-effects.h>
 #include <eel/eel-string.h>
@@ -55,8 +55,20 @@
 /* Should never be a reasonable actual mtime */
 #define INVALID_MTIME 0
 
+/* Cool-off period between last file modification time and thumbnail creation */
+#define THUMBNAIL_CREATION_DELAY_SECS 3
+
 static gpointer thumbnail_thread_start (gpointer data);
 
+/* structure used for making thumbnails, associating a uri with where the thumbnail is to be stored */
+
+typedef struct {
+	char *image_uri;
+	char *mime_type;
+	time_t original_file_mtime;
+} NautilusThumbnailInfo;
+
+
 /*
  * Thumbnail thread state.
  */
@@ -77,6 +89,9 @@ static volatile gboolean thumbnail_thread_is_running = FALSE;
 /* The list of NautilusThumbnailInfo structs containing information about the
    thumbnails we are making. Lock thumbnails_mutex when accessing this. */
 static volatile GList *thumbnails_to_make = NULL;
+/* The currently thumbnailed icon. it also exists in the thumbnails_to_make list
+ * to avoid adding it again. Lock thumbnails_mutex when accessing this. */
+static NautilusThumbnailInfo *currently_thumbnailing = NULL;
 
 static GnomeThumbnailFactory *thumbnail_factory = NULL;
 
@@ -99,15 +114,6 @@ get_file_mtime (const char *file_uri, time_t* mtime)
 	return TRUE;
 }
 
-
-/* structure used for making thumbnails, associating a uri with where the thumbnail is to be stored */
-
-typedef struct {
-	char *image_uri;
-	char *mime_type;
-	time_t original_file_mtime;
-} NautilusThumbnailInfo;
-
 /* GCompareFunc-style function for comparing NautilusThumbnailInfos.
  * Returns 0 if they refer to the same uri.
  */
@@ -123,6 +129,14 @@ compare_thumbnail_info (gconstpointer a, gconstpointer b)
 	return strcmp (info_a->image_uri, info_b->image_uri) != 0;
 }
 
+static void
+free_thumbnail_info (NautilusThumbnailInfo *info)
+{
+	g_free (info->image_uri);
+	g_free (info->mime_type);
+	g_free (info);
+}
+
 /* This function is added as a very low priority idle function to start the
    thread to create any needed thumbnails. It is added with a very low priority
    so that it doesn't delay showing the directory in the icon/list views.
@@ -143,7 +157,9 @@ thumbnail_thread_starter_cb (gpointer data)
 	pthread_attr_init (&thread_attributes);
 	pthread_attr_setdetachstate (&thread_attributes,
 				     PTHREAD_CREATE_DETACHED);
+#ifdef _POSIX_THREAD_ATTR_STACKSIZE
 	pthread_attr_setstacksize (&thread_attributes, 128*1024);
+#endif
 #ifdef DEBUG_THUMBNAILS
 	g_message ("(Main Thread) Creating thumbnails thread\n");
 #endif
@@ -211,57 +227,156 @@ nautilus_remove_thumbnail_for_file (const char *old_file_uri)
 	g_free (old_thumbnail_path);
 }
 
-/* routine to load an image from the passed-in path, and then embed it in
- * a frame if necessary
- */
-GdkPixbuf *
-nautilus_thumbnail_load_framed_image (const char *path)
+void
+nautilus_thumbnail_frame_image (GdkPixbuf **pixbuf)
 {
-	GdkPixbuf *pixbuf, *pixbuf_with_frame, *frame;
-	gboolean got_frame_offsets;
-	char *frame_offset_str;
+	GdkPixbuf *pixbuf_with_frame, *frame;
 	int left_offset, top_offset, right_offset, bottom_offset;
-	char c;
-	
-	pixbuf = gdk_pixbuf_new_from_file (path, NULL);
-	if (pixbuf == NULL) {
-		return NULL;
-	}
-	
+		
 	/* The pixbuf isn't already framed (i.e., it was not made by
 	 * an old Nautilus), so we must embed it in a frame.
 	 */
 
 	frame = nautilus_icon_factory_get_thumbnail_frame ();
 	if (frame == NULL) {
-		return pixbuf;
+		return;
 	}
 	
-	got_frame_offsets = FALSE;
-	frame_offset_str = nautilus_theme_get_theme_data ("thumbnails", "FRAME_OFFSETS");
-	if (frame_offset_str != NULL) {
-		if (sscanf (frame_offset_str, " %d , %d , %d , %d %c",
-			    &left_offset, &top_offset, &right_offset, &bottom_offset, &c) == 4) {
-			got_frame_offsets = TRUE;
-		}
-		g_free (frame_offset_str);
-	}
-	if (!got_frame_offsets) {
-		/* use nominal values since the info in the theme couldn't be found */
-		left_offset = 3;
-		top_offset = 3;
-		right_offset = 6;
-		bottom_offset = 6;
-	}
+	left_offset = 3;
+	top_offset = 3;
+	right_offset = 6;
+	bottom_offset = 6;
 	
 	pixbuf_with_frame = eel_embed_image_in_frame
-		(pixbuf, frame,
+		(*pixbuf, frame,
 		 left_offset, top_offset, right_offset, bottom_offset);
-	g_object_unref (pixbuf);	
+	g_object_unref (*pixbuf);	
 
-	return pixbuf_with_frame;
+	*pixbuf=pixbuf_with_frame;
 }
 
+/* routine to load an image from the passed-in path, and then embed it in
+ * a frame if necessary
+ */
+GdkPixbuf *
+nautilus_thumbnail_load_framed_image (const char *path)
+{
+	GdkPixbuf *pixbuf;
+	
+	pixbuf = gdk_pixbuf_new_from_file (path, NULL);
+	if (pixbuf == NULL) {
+		return NULL;
+	}
+        nautilus_thumbnail_frame_image(&pixbuf);	
+        return pixbuf;
+}
+
+void
+nautilus_thumbnail_remove_from_queue (const char *file_uri)
+{
+	NautilusThumbnailInfo info;
+	GList *node;
+	
+#ifdef DEBUG_THUMBNAILS
+	g_message ("(Remove from queue) Locking mutex\n");
+#endif
+	pthread_mutex_lock (&thumbnails_mutex);
+
+	/*********************************
+	 * MUTEX LOCKED
+	 *********************************/
+
+	info.image_uri = (char *)file_uri;
+	info.mime_type = NULL;
+	
+	node = g_list_find_custom ((GList*) thumbnails_to_make, &info,
+				   compare_thumbnail_info);
+
+	if (node && node->data != currently_thumbnailing) {
+		free_thumbnail_info (node->data);
+		thumbnails_to_make = g_list_delete_link ((GList *)thumbnails_to_make, node);
+	}
+	
+	/*********************************
+	 * MUTEX UNLOCKED
+	 *********************************/
+	
+#ifdef DEBUG_THUMBNAILS
+	g_message ("(Remove from queue) Unlocking mutex\n");
+#endif
+	pthread_mutex_unlock (&thumbnails_mutex);
+}
+
+void
+nautilus_thumbnail_remove_all_from_queue (void)
+{
+	GList *l, *next;
+	
+#ifdef DEBUG_THUMBNAILS
+	g_message ("(Remove all from queue) Locking mutex\n");
+#endif
+	pthread_mutex_lock (&thumbnails_mutex);
+
+	/*********************************
+	 * MUTEX LOCKED
+	 *********************************/
+
+	l = (GList *)thumbnails_to_make;
+	while (l != NULL) {
+		next = l->next;
+		if (l->data != currently_thumbnailing) {
+			free_thumbnail_info (l->data);
+			thumbnails_to_make = g_list_delete_link ((GList *)thumbnails_to_make, l);
+		}
+
+		l = next;
+	}
+	
+	/*********************************
+	 * MUTEX UNLOCKED
+	 *********************************/
+	
+#ifdef DEBUG_THUMBNAILS
+	g_message ("(Remove all from queue) Unlocking mutex\n");
+#endif
+	pthread_mutex_unlock (&thumbnails_mutex);
+}
+
+void
+nautilus_thumbnail_prioritize (const char *file_uri)
+{
+	NautilusThumbnailInfo info;
+	GList *node;
+
+#ifdef DEBUG_THUMBNAILS
+	g_message ("(Prioritize) Locking mutex\n");
+#endif
+	pthread_mutex_lock (&thumbnails_mutex);
+
+	/*********************************
+	 * MUTEX LOCKED
+	 *********************************/
+
+	info.image_uri = (char *)file_uri;
+	info.mime_type = NULL;
+	
+	node = g_list_find_custom ((GList*) thumbnails_to_make, &info,
+				   compare_thumbnail_info);
+
+	if (node && node->data != currently_thumbnailing) {
+		thumbnails_to_make = g_list_remove_link ((GList *)thumbnails_to_make, node);
+		thumbnails_to_make = g_list_concat (node, (GList *)thumbnails_to_make);
+	}
+	
+	/*********************************
+	 * MUTEX UNLOCKED
+	 *********************************/
+	
+#ifdef DEBUG_THUMBNAILS
+	g_message ("(Prioritize) Unlocking mutex\n");
+#endif
+	pthread_mutex_unlock (&thumbnails_mutex);
+}
 
 
 /***************************************************************************
@@ -286,6 +401,7 @@ thumbnail_thread_notify_file_changed (gpointer image_uri)
 #endif
 
 	if (file != NULL) {
+		nautilus_file_set_is_thumbnailing (file, FALSE);
 		nautilus_file_changed (file);
 		nautilus_file_unref (file);
 	}
@@ -301,6 +417,10 @@ nautilus_create_thumbnail (NautilusFile *file)
 {
 	time_t file_mtime = 0;
 	NautilusThumbnailInfo *info;
+	NautilusThumbnailInfo *existing_info;
+	GList *existing;
+
+	nautilus_file_set_is_thumbnailing (file, TRUE);
 
 	info = g_new0 (NautilusThumbnailInfo, 1);
 	info->image_uri = nautilus_file_get_uri (file);
@@ -329,15 +449,16 @@ nautilus_create_thumbnail (NautilusFile *file)
 	 *********************************/
 
 	/* Check if it is already in the list of thumbnails to make. */
-	if (g_list_find_custom ((GList*) thumbnails_to_make, info,
-				compare_thumbnail_info) == NULL) {
+	existing = g_list_find_custom ((GList*) thumbnails_to_make, info,
+				       compare_thumbnail_info);
+	if (existing == NULL) {
 		/* Add the thumbnail to the list. */
 #ifdef DEBUG_THUMBNAILS
 		g_message ("(Main Thread) Adding thumbnail: %s\n",
 			   info->image_uri);
 #endif
 		thumbnails_to_make = g_list_append ((GList*) thumbnails_to_make, info);
-
+		
 		/* If the thumbnail thread isn't running, and we haven't
 		   scheduled an idle function to start it up, do that now.
 		   We don't want to start it until all the other work is done,
@@ -347,11 +468,17 @@ nautilus_create_thumbnail (NautilusFile *file)
 			thumbnail_thread_starter_id = g_idle_add_full (G_PRIORITY_LOW, thumbnail_thread_starter_cb, NULL, NULL);
 		}
 	} else {
-		g_free (info->image_uri);
-		g_free (info->mime_type);
-		g_free (info);
+#ifdef DEBUG_THUMBNAILS
+		g_message ("(Main Thread) Updating non-current mtime: %s\n",
+			   info->image_uri);
+#endif
+		/* The file in the queue might need a new original mtime */
+		existing_info = existing->data;
+		existing_info->original_file_mtime = info->original_file_mtime;
+		free_thumbnail_info (info);
 	}
-	
+	    
+
 	/*********************************
 	 * MUTEX UNLOCKED
 	 *********************************/
@@ -368,6 +495,8 @@ thumbnail_thread_start (gpointer data)
 {
 	NautilusThumbnailInfo *info = NULL;
 	GdkPixbuf *pixbuf;
+	time_t current_orig_mtime = 0;
+	time_t current_time;
 
 	/* We loop until there are no more thumbails to make, at which point
 	   we exit the thread. */
@@ -384,13 +513,17 @@ thumbnail_thread_start (gpointer data)
 		/* Pop the last thumbnail we just made off the head of the
 		   list and free it. I did this here so we only have to lock
 		   the mutex once per thumbnail, rather than once before
-		   creating it and once after. */
-		if (thumbnails_to_make && info == thumbnails_to_make->data) {
-			thumbnails_to_make = g_list_remove ((GList*) thumbnails_to_make, info);
-			g_free (info->image_uri);
-			g_free (info->mime_type);
-			g_free (info);
+		   creating it and once after.
+		   Don't pop the thumbnail off the queue if the original file
+		   mtime of the request changed. Then we need to redo the thumbnail.
+		*/
+		if (currently_thumbnailing &&
+		    currently_thumbnailing->original_file_mtime == current_orig_mtime) {
+			g_assert (info == currently_thumbnailing);
+			free_thumbnail_info (currently_thumbnailing);
+			thumbnails_to_make = g_list_remove ((GList*) thumbnails_to_make, currently_thumbnailing);
 		}
+		currently_thumbnailing = NULL;
 
 		/* If there are no more thumbnails to make, reset the
 		   thumbnail_thread_is_running flag, unlock the mutex, and
@@ -408,7 +541,8 @@ thumbnail_thread_start (gpointer data)
 		   is created so the main thread doesn't add it again while we
 		   are creating it. */
 		info = thumbnails_to_make->data;
-
+		currently_thumbnailing = info;
+		current_orig_mtime = info->original_file_mtime;
 		/*********************************
 		 * MUTEX UNLOCKED
 		 *********************************/
@@ -418,6 +552,21 @@ thumbnail_thread_start (gpointer data)
 #endif
 		pthread_mutex_unlock (&thumbnails_mutex);
 
+		time (&current_time);
+
+		/* Don't try to create a thumbnail if the file was modified recently.
+		   This prevents constant re-thumbnailing of changing files. */ 
+		if (current_time < current_orig_mtime + THUMBNAIL_CREATION_DELAY_SECS &&
+		    current_time >= current_orig_mtime) {
+#ifdef DEBUG_THUMBNAILS
+			g_message ("(Thumbnail Thread) Skipping: %s\n",
+				   info->image_uri);
+#endif
+			/* Reschedule thumbnailing via a change notification */
+			g_timeout_add (1000, thumbnail_thread_notify_file_changed,
+				       g_strdup (info->image_uri));
+ 			continue;
+		}
 
 		/* Create the thumbnail. */
 #ifdef DEBUG_THUMBNAILS
@@ -433,12 +582,12 @@ thumbnail_thread_start (gpointer data)
 			gnome_thumbnail_factory_save_thumbnail (thumbnail_factory,
 								pixbuf,
 								info->image_uri,
-								info->original_file_mtime);
+								current_orig_mtime);
 			g_object_unref (pixbuf);
 		} else {
 			gnome_thumbnail_factory_create_failed_thumbnail (thumbnail_factory, 
 									 info->image_uri,
-									 info->original_file_mtime);
+									 current_orig_mtime);
 		}
 		/* We need to call nautilus_file_changed(), but I don't think that is
 		   thread safe. So add an idle handler and do it from the main loop. */
